#include <XSUtil/Utils/XSutility.h>
#include <XSUtil/FunctionUtils/FunctionUtility.h>
#include <XSUtil/FunctionUtils/xsFortran.h>
#include <fitsio.h>
#include <sstream>
#include <algorithm>

namespace {
   void errReturn(fitsfile* contfil, fitsfile* linefil, 
                const string& context, int status);
}


typedef XSutility::auto_array_ptr<float> FloatAP;
typedef XSutility::auto_array_ptr<int> IntAP;


extern "C" void rdneiaped_(int& iocont, int& ioline, int& nmtval, int& nmions, 
                   int& ncont, int& npseudo, int& nlines, float* tval, float* cont, 
                   float* cenerg, int* conEl, int* conIon, int* conpt, int* consiz, 
                   float* pseudo, float* penerg, int* psupt, int* psusiz, 
                   float* lineE, float* lineEm, int* lineEl, int* lineIon,
                   int* linpt, int* linsiz, int* status);

void ldneiaped(FloatAP& tval, long& nmtval, long& nmions, FloatAP& cont, FloatAP& cenerg,
        IntAP& conEl, IntAP& conIon, IntAP& conpt, IntAP& consiz, FloatAP& pseudo,
        FloatAP& penerg, IntAP& psupt, IntAP& psusiz, FloatAP& lineE, FloatAP& lineEm,
        IntAP& lineEl, IntAP& lineIon, IntAP& linpt, IntAP& linsiz, int& status)
{
   // translated from ldneiaped.f by C.Gordon 01/09

   // Subroutine to read the NEI APED lasma model files.
   // Arguments :
   //     tval             r: pointer to array of tabulated temperatures
   //     nmtval           r: number of tabulated temperatures
   //     nmions           r: max number of tabulated ions per temperature
   //     cont             r: Tabulated continua
   //     cenerg           r: Energies for tabulated continua
   //     conEl            r: Atomic numbers for tabulated continua
   //     conIon           r: Ion stages for tabulated continua
   //     conpt            r: Offsets in the cont array for each ion and kT
   //     consiz           r: Sizes in the cont array for each ion and kT
   //     pseudo           r: Tabulated pseudo-continua
   //     penerg           r: Energies for tabulated pseudo-continua
   //     psupt            r: Offsets in the pseudo array for each ion and kT
   //     psusiz           r: Sizes in the pseudo array for each ion and kT
   //     lineE            r: Tabulated line energies
   //     lineEm           r: Tabulated line emissivities
   //     lineEl           r: Tabulated line elements
   //     lineIon          r: Tabulated line ions
   //     linpt            r: Offsets in the line arrays for each kT
   //     linsiz           r: Sizes in the line arrays for each kT
   
   status=0;

   // Set the names of the input files and open them. Note that for APEC
   // we check for the NEIAPECROOT string model parameter to allow the user
   // to substitute different _coco and _line tables.

   const string& datadir = FunctionUtility::modelDataPath();
   string pname("NEIAPECROOT");
   string pvalue(FunctionUtility::getModelString(pname));
   string contfil;
   string linefil;
   int anynul=0, hdutype=0;
   static string ovalue;
   
   if (!pvalue.length() || pvalue == FunctionUtility::NOT_A_KEY())
   {
      contfil = datadir + "APEC_nei_v11_comp.fits";
      linefil = datadir + "APEC_nei_v11_line.fits";
   }
   else
   {
      contfil = pvalue + "_comp.fits";
      linefil = pvalue + "_line.fits";
      if (ovalue != pvalue)
      {
         ovalue = pvalue;
         string msg("Using NEIAPEC root filename " + pvalue);
         xs_write(const_cast<char*>(msg.c_str()), 10);
      }
   }
   
   fitsfile *pContFil=0;
   fits_open_file(&pContFil, const_cast<char*>(contfil.c_str()), 0, &status);
   if (status)
   {
      std::ostringstream err;
      err << "Failed to open " << contfil <<"\nError in LDNEIAPED : status = "
          << status;
      xs_write(const_cast<char*>(err.str().c_str()), 10);
      return;
   }
   
   fitsfile *pLineFil=0;
   fits_open_file(&pLineFil, const_cast<char*>(linefil.c_str()), 0, &status);
   if (status)
   {
      std::ostringstream err;
      err << "Failed to open " << linefil <<"\nError in LDNEIAPED : status = "
          << status;
      xs_write(const_cast<char*>(err.str().c_str()), 10);
      int status2=0;
      fits_close_file(pContFil, &status2);
      return;
   } 
   
   // Read the PARAMETERS extension to find the sizes of arrays that will be 
   // required. First the continuum data

   if (fits_movnam_hdu(pContFil, BINARY_TBL, "PARAMETERS", 0, &status))
   {
      string context("Failed to find PARAMETERS extension in continuum file");
      errReturn(pContFil, pLineFil, context, status);
      return;
   }  

   // Get the number of temperatures

   if (fits_read_key_lng(pContFil, "NAXIS2", &nmtval, 0, &status))
   {
      string context("Failed to read NAXIS2 in continuum PARAMETERS extension");
      errReturn(pContFil, pLineFil, context, status);
      return;      
   }

   // and the maximum number of ions per temperature

   nmions = 0;
   int icol=0;
   if (fits_get_colnum(pContFil, CASEINSEN, "NElement", &icol, &status))
   {
      string context("Failed to find NElement col in continuum PARAMETERS extension");
      errReturn(pContFil, pLineFil, context, status);
      return;      
   }
   
   long longBuf=0;
   for (long i=1; i<=nmtval; ++i)
   {
      if (fits_read_col_lng(pContFil, icol, i, 1, 1, 0, &longBuf, &anynul, &status))
      {
         std::ostringstream oss;
         oss << "Failed to read NElement : " << i;
         string context(oss.str());
         errReturn(pContFil, pLineFil, context, status);
         return;      
      }
      nmions = std::max(nmions, longBuf);
   }

   // Sum up the total continuum and pseudo-continuum elements required. This
   // requires looping round the emissivity extensions.

   int ncont=0;
   int npseudo=0;
   
   for (long i=1; i<=nmtval; ++i)
   {
      if (fits_movrel_hdu(pContFil, 1, &hdutype, &status))
      {
         std::ostringstream oss;
         oss << "Failed to move to extension for temperature # " << i;
         string context(oss.str());
         errReturn(pContFil, pLineFil, context, status);
         return;
      }
      long nrows=0;
      if (fits_read_key_lng(pContFil, "NAXIS2", &nrows, 0, &status))
      {
         std::ostringstream oss;
         oss << "Failed to read NAXIS2 for temperature # " << i;
         string context(oss.str());
         errReturn(pContFil, pLineFil, context, status);
         return;
      }
      if (fits_get_colnum(pContFil, CASEINSEN, "N_Cont", &icol, &status))
      {
         std::ostringstream oss;
         oss << "Failed find N_Cont column for temperature # " << i;
         string context(oss.str());
         errReturn(pContFil, pLineFil, context, status);
         return;
      }
      for (long j=1; j<=nrows; ++j)
      {
         if (fits_read_col_lng(pContFil, icol, j, 1, 1, 0, &longBuf, &anynul, &status))
         {
            std::ostringstream oss;
            oss << "Failed to read N_Cont : " << i << " " << j;
            string context(oss.str());
            errReturn(pContFil, pLineFil, context, status);
            return;
         }
         ncont += static_cast<int>(longBuf);
      }
      if (fits_get_colnum(pContFil, CASEINSEN, "N_Pseudo", &icol, &status))
      {
         std::ostringstream oss;
         oss << "Failed find N_Pseudo column for temperature # " << i;
         string context(oss.str());
         errReturn(pContFil, pLineFil, context, status);
         return;
      }
      for (long j=1; j<=nrows; ++j)
      {
         if (fits_read_col_lng(pContFil, icol, j, 1, 1, 0, &longBuf, &anynul, &status))
         {
            std::ostringstream oss;
            oss << "Failed to read N_Pseudo : " << i << " " << j;
            string context(oss.str());
            errReturn(pContFil, pLineFil, context, status);
            return;
         }
         npseudo += static_cast<int>(longBuf);
      }
   } // end tval loop
   
   // Now the line data
   
   if (fits_movnam_hdu(pLineFil, BINARY_TBL, "PARAMETERS", 0, &status))
   {
      string context("Failed to find PARAMETERS extension in line file");
      errReturn(pContFil, pLineFil, context, status);
      return;
   }  
   
   int nlines=0;
   
   if (fits_get_colnum(pLineFil, CASEINSEN, "Nline", &icol, &status))
   {
      string context("Failed to find Nline column in line PARAMETERS extension");
      errReturn(pContFil, pLineFil, context, status);
      return;
   }
   for (long i=1; i<=nmtval; ++i)
   {
      if (fits_read_col_lng(pLineFil, icol, i, 1, 1, 0, &longBuf, &anynul, &status))
      {
         std::ostringstream oss;
         oss << "Failed to read Nline : " << i;
         string context(oss.str());
         errReturn(pContFil, pLineFil, context, status);
         return;      
      }
      nlines += static_cast<int>(longBuf);
   }
   
   // Reset both files to the primary HDU
   
   if (fits_movabs_hdu(pContFil, 1, &hdutype, &status))
   {
      string context("Failed to move back to continuum file primary header");
      errReturn(pContFil, pLineFil, context, status);
      return;
   }
   if (fits_movabs_hdu(pLineFil, 1, &hdutype, &status))
   {
      string context("Failed to move back to line file primary header");
      errReturn(pContFil, pLineFil, context, status);
      return;
   }
   
   // Grab the memory that will be needed
   
   float *oldF = tval.reset(new float[nmtval]);
   delete [] oldF;
   oldF = cont.reset(new float[ncont]);
   delete [] oldF;
   oldF = cenerg.reset(new float[ncont]);
   delete [] oldF;
   int *oldI = conEl.reset(new int[nmtval*nmions]);
   delete [] oldI;
   oldI = conIon.reset(new int[nmtval*nmions]);
   delete [] oldI;
   oldI = conpt.reset(new int[nmtval*nmions]);
   delete [] oldI;
   oldI = consiz.reset(new int[nmtval*nmions]);
   delete [] oldI;
   oldF = pseudo.reset(new float[npseudo]);
   delete [] oldF;
   oldF = penerg.reset(new float[npseudo]);
   delete [] oldF;
   oldI = psupt.reset(new int[nmtval*nmions]);
   delete [] oldI;
   oldI = psusiz.reset(new int[nmtval*nmions]);
   delete [] oldI;
   oldF = lineE.reset(new float[nlines]);
   delete [] oldF;
   oldF = lineEm.reset(new float[nlines]);
   delete [] oldF;
   oldI = lineEl.reset(new int[nlines]);
   delete [] oldI;
   oldI = lineIon.reset(new int[nlines]);
   delete [] oldI;
   oldI = linpt.reset(new int[nmtval]);
   delete [] oldI;
   oldI = linsiz.reset(new int[nmtval]);
   delete [] oldI;
   
   // Read the files and load the arrays
   
   int iocont = CFITS2Unit(pContFil);
   int ioline = CFITS2Unit(pLineFil);
   int intnmtval = static_cast<int>(nmtval);
   int intnmions = static_cast<int>(nmions);
   rdneiaped_(iocont, ioline, intnmtval, intnmions, ncont, npseudo, nlines,
      tval.get(), cont.get(), cenerg.get(), conEl.get(), conIon.get(),
      conpt.get(), consiz.get(), pseudo.get(), penerg.get(), psupt.get(),
      psusiz.get(), lineE.get(), lineEm.get(), lineEl.get(), lineIon.get(),
      linpt.get(), linsiz.get(), &status);
     
   if (status)
   {
      string context("Failure in RDNEIAPED");
      errReturn(pContFil, pLineFil, context, status);
      return;
   }
   
   // Close the files
   
   fits_close_file(pContFil, &status);
   if (status)
   {
      std::ostringstream oss;
      oss << "Failed to close continuum file: Error in LDNEIAPED: "
          << "status = " << status;
      string errMsg(oss.str());
      xs_write(const_cast<char*>(errMsg.c_str()), 10);
      return;
   }

   fits_close_file(pLineFil, &status);
   if (status)
   {
      std::ostringstream oss;
      oss << "Failed to close line file: Error in LDNEIAPED: "
          << "status = " << status;
      string errMsg(oss.str());
      xs_write(const_cast<char*>(errMsg.c_str()), 10);
      return;
   }
}


namespace {
   void errReturn(fitsfile* contfil, fitsfile* linefil, 
                const string& context, int status)
   {
      std::ostringstream oss;
      oss << context << "\nError in LDNEIAPED : status = " << status;
      string fullMsg(oss.str());
      xs_write(const_cast<char*>(fullMsg.c_str()), 10);
      int status2=0;
      fits_close_file(contfil, &status2);
      fits_close_file(linefil, &status2);
   }
}
